#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
# Univention Management Console module:
#   Manage virtual classrooms
#
# Copyright 2014 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import re
import copy
import json
from subprocess import Popen, PIPE
from tempfile import NamedTemporaryFile

from univention.admin.uldap import getMachineConnection
from ucsschool.lib.models import School as SchoolModel

from univention.lib.i18n import Translation
from univention.config_registry.frontend import ucr_update
from univention.management.console.modules import UMC_Error as RequestError, Base
from univention.management.console.modules.decorators import simple_response, sanitize
from univention.management.console.modules.sanitizers import StringSanitizer, BooleanSanitizer, EmailSanitizer
from univention.management.console.log import MODULE
from univention.management.console.config import ucr

_ = Translation('univention-management-console-module-webweaver').translate

RE_SCHOOLS = re.compile(r'^webweaver/(?P<school>[^/]+)/(?P<type>ServerURL|ServerLogin|ServerPasswd|error_email)$')
RESYNC_COMMAND = '/usr/sbin/univention-digionline-webweaver'


class School(dict):

	def __init__(self, school, display_name, props=None):
		dict.__init__(self)
		self.update({
			'school': school,
			'displayName': display_name,
			'ServerURL': None,
			'ServerLogin': None,
			'ServerPasswd': None,
			'error_email': None,
		})
		if props:
			self.update(props)

	def get(self, key, default=None):
		return dict.get(self, key, {})


class Schools(dict):

	def __init__(self, lo):
		dict.__init__(self)
		self.load_ucr()
		schools = SchoolModel.get_all(lo)
		for school in schools:
			self[school.name] = School(school.name, school.display_name, self.schools.get(school.name))

	def load_ucr(self):
		ucr.load()
		schools = {}
		for key, value in ucr.items():
			match = RE_SCHOOLS.match(key)
			if match is None:
				continue
			match = match.groupdict()
			school, type_ = match['school'], match['type']
			schools.setdefault(school, {
				'ServerURL': None,
				'ServerLogin': None,
				'ServerPasswd': None,
				'error_email': None,
			})
			if type_ == 'error_email':
				value = ucr.is_true(None, None, value)
			schools[school][type_] = value
			schools[school]['school'] = school
		self.schools = schools

	def save_ucr(self, schools):
		if schools is None:
			schools = self
		new_entries = {}
		# remove old values
		for key in ucr.keys():
			if RE_SCHOOLS.match(key):
				new_entries[key] = None
		# add new values
		for school in schools.values():
			name = school['school']
			for type_, value in school.iteritems():
				if type_ == 'school':
					continue
				if isinstance(value, bool):
					value = 'true' if value else 'false'
				new_entries['webweaver/%s/%s' % (name, type_)] = value

		# commit changes to UCR
		ucr_update(ucr, new_entries)

	def put_school(self, school):
		schools = copy.deepcopy(self)
		name = school['school']
		school2 = self.get_school(name)
		school2.update(school)
		schools[name] = school2
		self.save_ucr(schools)

	def get_school(self, school):
		return self.get(school, School(school, school))

	def delete_school(self, school):
		schools = copy.deepcopy(self)
		try:
			del schools[school]
		except KeyError:
			raise RequestError(_('The specified school does not exists.'))
		else:
			self.save_ucr(schools)


class Instance(Base):

	def init(self):
		self.lo, self.po = getMachineConnection()
		self.load()

	def load(self):
		ucr.load()
		self.schools = Schools(self.lo)

	@simple_response
	def query(self):
		self.load()
		return [dict(school) for school in self.schools.values()]

	@simple_response
	def get(self, school):
		self.load()
		return self.schools.get_school(school)

	@sanitize(
		school=StringSanitizer(required=True),
		ServerLogin=EmailSanitizer(required=True),
		ServerPasswd=StringSanitizer(required=True),
		ServerURL=StringSanitizer(required=True),
		error_email=BooleanSanitizer(required=True)
	)
	@simple_response
	def put(self, school, ServerLogin, ServerPasswd, ServerURL, error_email):
		self.load()

		self.prevent_duplicated_username(ServerLogin, school)
		self.check_login_credentials(ServerURL, ServerLogin, ServerPasswd)

		self.schools.put_school({
			'school': school,
			'ServerURL': ServerURL,
			'ServerLogin': ServerLogin,
			'ServerPasswd': ServerPasswd,
			'error_email': error_email,
		})

	def prevent_duplicated_username(self, ServerLogin, school):
		# prevent duplicated username
		for entry in self.schools.values():
			if ServerLogin == entry['ServerLogin'] and entry['school'] != school:
				raise RequestError(_('The server login credentials are already used by the school "%s".') % (entry['school'],))

	def check_login_credentials(self, ServerURL, ServerLogin, ServerPasswd):
		try:
			client = WebWeaverSOAP(ServerURL, ServerLogin, ServerPasswd)
			client.test_credentials()
		except WebWeaverSOAP.Error as exc:
			MODULE.error('Credentials are wrong; PHP reports: returncode=%d; stdout=%r; stderr=%r' % (exc.returncode, exc.stdout, exc.stderr))
			raise RequestError(str(exc))
		else:
			pass

	@simple_response
	def delete(self, school):
		self.load()
		self.schools.delete_school(school)

	@simple_response
	def resync(self, school):
		self.load()
		try:
			self.schools[school]
		except KeyError:
			raise RequestError(_('The specified school does not exists.'))
		else:
			process = Popen([RESYNC_COMMAND, '--resync', school], stdout=PIPE, stderr=PIPE)
			stdout, stderr = process.communicate()

			if process.returncode != 0:
				raise RequestError('%s\n%s' % (_('Failed to resynchronize settings.'), stdout))
			return True


class WebWeaverSOAP(object):

	LOGIN_SCRIPT = '/usr/share/univention-digionline-webweaver/test_credentials.php'

	class Error(Exception):
		def __init__(self, returncode, stdout, stderr):
			self.returncode = returncode
			self.stdout = stdout
			self.stderr = stderr

		def __str__(self):
			msg = str(_('The specified credentials are invalid.'))
			if self.stdout:
				msg = '%s\n%s\n%s' % (msg, _('Server responded with:'), self.stdout)
			return msg

	def __init__(self, server_url, server_login, server_password):
		self.server_url = server_url
		self.server_login = server_login
		self.server_password = server_password

	def test_credentials(self):
		fd = self._create_credentials_file()
		try:
			process = Popen([self.LOGIN_SCRIPT, fd.name], stdout=PIPE, stderr=PIPE)
			stdout, stderr = process.communicate()
			if process.returncode:
				raise self.Error(process.returncode, stdout.strip(), stderr.strip())
			return True
		finally:
			# remove tempfile
			fd.close()

	def _create_credentials_file(self):
		credentials = dict(
			ServerLogin=self.server_login,
			ServerURL=self.server_url,
			ServerPasswd=self.server_password
		)
		fd = NamedTemporaryFile()
		json.dump(credentials, fd)
		fd.flush()
		return fd
